import numpy as np
from pytest_cases import parametrize_with_cases
import xarray as xr
from pytest import raises
import pandas as pd

from tests.conftest import _compare_metrics_df_with_xarray
from gval.utils.loading_datasets import (
    adjust_memory_strategy,
    get_current_memory_strategy,
    _handle_xarray_memory,
)


@parametrize_with_cases(
    "candidate_map, benchmark_map, positive_categories, negative_categories, rasterize_attributes, memory_strategies, comparison_function",
    glob="data_array_accessor_success",
)
def test_data_array_accessor_success(
    candidate_map,
    benchmark_map,
    positive_categories,
    negative_categories,
    rasterize_attributes,
    memory_strategies,
    comparison_function,
):
    adjust_memory_strategy(memory_strategies)

    assert get_current_memory_strategy() == memory_strategies

    candidate_map = _handle_xarray_memory(candidate_map)

    data = candidate_map.gval.categorical_compare(
        benchmark_map=benchmark_map,
        positive_categories=positive_categories,
        negative_categories=negative_categories,
        rasterize_attributes=rasterize_attributes,
        comparison_function=comparison_function,
        allow_candidate_values=[1, 2, np.nan],
        allow_benchmark_values=[0, 2, np.nan],
    )

    if rasterize_attributes is not None:
        assert isinstance(data[0], pd.DataFrame)
    else:
        assert isinstance(data[0], xr.DataArray)
    assert isinstance(data[1], pd.DataFrame)
    assert isinstance(data[2], pd.DataFrame)


@parametrize_with_cases(
    "candidate_map, benchmark_map, positive_categories, negative_categories, memory_strategies, exception",
    glob="data_array_accessor_fail",
)
def test_data_array_accessor_fail(
    candidate_map,
    benchmark_map,
    positive_categories,
    negative_categories,
    memory_strategies,
    exception,
):
    with raises(exception):
        adjust_memory_strategy(memory_strategies)

        if exception == OSError:
            candidate_map.encoding["source"] = "arb"
            _handle_xarray_memory(candidate_map)

        _ = candidate_map.gval.categorical_compare(
            benchmark_map=benchmark_map,
            positive_categories=positive_categories,
            negative_categories=negative_categories,
        )


@parametrize_with_cases(
    "candidate_map, benchmark_map, rasterize_attributes",
    glob="data_array_accessor_homogenize",
)
def test_data_array_accessor_homogenize(
    candidate_map, benchmark_map, rasterize_attributes
):
    data = candidate_map.gval.homogenize(
        benchmark_map=benchmark_map, rasterize_attributes=rasterize_attributes
    )

    assert isinstance(data[0], xr.DataArray)
    assert isinstance(data[1], xr.DataArray)


@parametrize_with_cases(
    "candidate_map, benchmark_map, vectorized",
    glob="data_array_accessor_compute_agreement",
)
def test_data_array_accessor_compute_agreement_success(
    candidate_map, benchmark_map, vectorized
):
    aligned_cand, aligned_bench = candidate_map.gval.homogenize(
        benchmark_map=benchmark_map
    )

    if vectorized:
        aligned_cand.gval.agreement_map_format = "vector"

    data = aligned_cand.gval.compute_agreement_map(benchmark_map=aligned_bench)

    assert isinstance(data, xr.DataArray)


@parametrize_with_cases(
    "candidate_map, benchmark_map", glob="data_array_accessor_compute_agreement"
)
def test_data_array_accessor_compute_agreement_fail(candidate_map, benchmark_map):
    with raises(ValueError):
        _ = candidate_map.gval.compute_agreement_map(benchmark_map=benchmark_map)


@parametrize_with_cases(
    "agreement_map",
    glob="data_array_accessor_crosstab_table_success",
)
def test_data_array_accessor_crosstab_table_success(agreement_map):
    data = agreement_map.gval.compute_crosstab()

    assert isinstance(data, pd.DataFrame)


@parametrize_with_cases(
    "agreement_map, exception",
    glob="data_array_accessor_crosstab_table_fail",
)
def test_data_array_accessor_crosstab_table_fail(agreement_map, exception):
    with raises(exception):
        _ = agreement_map.gval.compute_crosstab()


@parametrize_with_cases(
    "candidate_map, benchmark_map, positive_categories, negative_categories, rasterize_attributes",
    glob="data_set_accessor_success",
)
def test_data_set_accessor_success(
    candidate_map,
    benchmark_map,
    positive_categories,
    negative_categories,
    rasterize_attributes,
):
    data = candidate_map.gval.categorical_compare(
        benchmark_map=benchmark_map,
        positive_categories=positive_categories,
        negative_categories=negative_categories,
        rasterize_attributes=rasterize_attributes,
    )

    if rasterize_attributes is None:
        assert isinstance(data[0], xr.Dataset)
    else:
        assert isinstance(data[0], pd.DataFrame)
    assert isinstance(data[1], pd.DataFrame)
    assert isinstance(data[2], pd.DataFrame)


@parametrize_with_cases(
    "candidate_map, benchmark_map, rasterize_attributes",
    glob="data_set_accessor_homogenize",
)
def test_data_set_accessor_homogenize(
    candidate_map, benchmark_map, rasterize_attributes
):
    data = candidate_map.gval.homogenize(
        benchmark_map=benchmark_map, rasterize_attributes=rasterize_attributes
    )

    assert isinstance(data[0], xr.Dataset)
    assert isinstance(data[1], xr.Dataset)


@parametrize_with_cases(
    "candidate_map, benchmark_map", glob="data_set_accessor_compute_agreement"
)
def test_data_set_accessor_compute_agreement_success(candidate_map, benchmark_map):
    aligned_cand, aligned_bench = candidate_map.gval.homogenize(
        benchmark_map=benchmark_map
    )
    data = aligned_cand.gval.compute_agreement_map(benchmark_map=aligned_bench)

    assert isinstance(data, xr.Dataset)


@parametrize_with_cases(
    "candidate_map, benchmark_map", glob="data_set_accessor_compute_agreement"
)
def test_data_set_accessor_compute_agreement_fail(candidate_map, benchmark_map):
    with raises(ValueError):
        _ = candidate_map.gval.compute_agreement_map(benchmark_map=benchmark_map)


@parametrize_with_cases(
    "agreement_map", glob="data_set_accessor_crosstab_table_success"
)
def test_data_set_accessor_crosstab_table_success(agreement_map):
    data = agreement_map.gval.compute_crosstab()

    assert isinstance(data, pd.DataFrame)


@parametrize_with_cases(
    "agreement_map",
    glob="data_set_accessor_crosstab_table_fail",
)
def test_data_set_accessor_crosstab_table_fail(agreement_map):
    with raises(KeyError):
        _ = agreement_map.gval.compute_crosstab()


@parametrize_with_cases(
    "crosstab_df, positive_categories, negative_categories",
    glob="data_frame_accessor_compute_metrics",
)
def test_data_frame_accessor_compute_metrics(
    crosstab_df, positive_categories, negative_categories
):
    data = crosstab_df.gval.compute_categorical_metrics(
        positive_categories=positive_categories, negative_categories=negative_categories
    )

    assert isinstance(data, pd.DataFrame)


@parametrize_with_cases(
    "candidate_map, benchmark_map",
    glob="data_array_accessor_continuous",
)
def test_data_array_accessor_continuous(candidate_map, benchmark_map):
    agreement_map, metrics_df = candidate_map.gval.continuous_compare(
        benchmark_map=benchmark_map, metrics="all"
    )

    assert isinstance(agreement_map, xr.DataArray)
    assert isinstance(metrics_df, pd.DataFrame)


@parametrize_with_cases(
    "candidate_map, benchmark_map",
    glob="data_set_accessor_continuous",
)
def test_data_set_accessor_continuous(
    candidate_map,
    benchmark_map,
):
    agreement_map, metrics_df = candidate_map.gval.continuous_compare(
        benchmark_map=benchmark_map, metrics="all"
    )

    assert isinstance(agreement_map, xr.Dataset)
    assert isinstance(metrics_df, pd.DataFrame)


@parametrize_with_cases(
    "candidate_map, benchmark_map, agreement_map",
    glob="accessor_attributes",
)
def test_accessor_attributes(candidate_map, benchmark_map, agreement_map):
    attrs_df, agreement_map = candidate_map.gval.attribute_tracking(
        benchmark_map=benchmark_map, agreement_map=agreement_map
    )

    assert isinstance(agreement_map, xr.DataArray)
    assert isinstance(attrs_df, pd.DataFrame)


@parametrize_with_cases(
    "vector_map, reference_map, attributes",
    glob="dataframe_accessor_rasterize",
)
def test_dataframe_accessor_rasterize(vector_map, reference_map, attributes):
    raster_map = vector_map.gval.rasterize_data(
        reference_map=reference_map, rasterize_attributes=attributes
    )

    assert isinstance(raster_map, type(reference_map))
    assert raster_map.shape == reference_map.shape


@parametrize_with_cases(
    "candidate_map, benchmark_map, compute_kwargs, expected_df",
    glob="data_array_accessor_probabilistic_success",
)
def test_data_array_accessor_probabilistic_success(
    candidate_map, benchmark_map, compute_kwargs, expected_df
):
    _, metrics_df = candidate_map.gval.probabilistic_compare(
        benchmark_map, **compute_kwargs
    )

    _compare_metrics_df_with_xarray(metrics_df, expected_df)



"""
Tests for the catalog module
"""

from pytest_cases import parametrize_with_cases
from pytest import raises

import pandas as pd
import dask.dataframe as dd
import rioxarray as rxr
import xarray as xr
import pystac_client

from tests.conftest import _attributes_to_string
from gval.catalogs.catalogs import catalog_compare
from gval.utils.loading_datasets import stac_to_df, _parse_string_attributes


@parametrize_with_cases(
    "candidate_catalog, benchmark_catalog, on, map_ids, how, compare_type, compare_kwargs, open_kwargs, agreement_map_field, expected, expected_agreement_map",
    glob="compare_catalogs",
)
def test_compare_catalogs(
    candidate_catalog,
    benchmark_catalog,
    on,
    map_ids,
    how,
    compare_type,
    compare_kwargs,
    open_kwargs,
    agreement_map_field,
    expected,
    expected_agreement_map,
    tmp_path,
):
    """Tests for the compare_catalogs function"""

    def prepend_tmp_path(df, field):
        """Prepend tmp_path to field in df"""
        if (field is not None) & (field in df.columns):

            def tmp_path_apply(row):
                return tmp_path / row[field]

            dask_meta = (
                {"meta": ("output", "O")} if isinstance(df, dd.DataFrame) else {}
            )
            df[field] = df.apply(tmp_path_apply, axis=1, **dask_meta)
        return df

    # apply tmp_path to agreement_map_field in all catalogs
    candidate_catalog = prepend_tmp_path(candidate_catalog, agreement_map_field)
    benchmark_catalog = prepend_tmp_path(benchmark_catalog, agreement_map_field)
    expected = prepend_tmp_path(expected, agreement_map_field)

    agreement_catalog = catalog_compare(
        candidate_catalog,
        benchmark_catalog,
        on=on,
        map_ids=map_ids,
        how=how,
        compare_type=compare_type,
        compare_kwargs=compare_kwargs,
        open_kwargs=open_kwargs,
        agreement_map_field=agreement_map_field,
    )

    if isinstance(agreement_catalog, dd.DataFrame):
        agreement_catalog = agreement_catalog.compute().reset_index(drop=True)

    # check that the columns are the same without order
    pd.testing.assert_index_equal(
        agreement_catalog.columns, expected.columns, check_order=False
    )

    # check that the columns are the same with order. if not, reorder
    try:
        pd.testing.assert_index_equal(
            agreement_catalog.columns, expected.columns, check_order=True
        )
    except AssertionError:
        # make the same order of columns
        agreement_catalog = agreement_catalog[expected.columns]
        pd.testing.assert_index_equal(
            agreement_catalog.columns, expected.columns, check_order=True
        )

    # check dtypes. if not the same, recast
    agreement_catalog = agreement_catalog.astype(expected.dtypes)
    pd.testing.assert_series_equal(agreement_catalog.dtypes, expected.dtypes)
    agreement_catalog["agreement_maps"] = agreement_catalog["agreement_maps"].astype(
        str
    )
    expected["agreement_maps"] = expected["agreement_maps"].astype(str)

    pd.testing.assert_frame_equal(
        agreement_catalog, expected, check_dtype=False, check_like=True
    )

    # load agreement maps and check metadata
    if agreement_map_field is not None:
        # load agreement maps with apply and check that they are the same

        def load_agreement_and_check(row, counter=[0]):
            """Load agreement map and check that it is the same"""
            agreement_map = rxr.open_rasterio(row[agreement_map_field], **open_kwargs)

            if expected_agreement_map[counter[0]] is not None:
                expected_agreement_map_xr = rxr.open_rasterio(
                    expected_agreement_map[counter[0]], **open_kwargs
                )

                xr.testing.assert_identical(
                    _attributes_to_string(_parse_string_attributes(agreement_map)),
                    _attributes_to_string(expected_agreement_map_xr),
                )

            # increment counter
            counter[0] += 1

        _ = agreement_catalog.apply(load_agreement_and_check, axis=1)


@parametrize_with_cases(
    "candidate_catalog, benchmark_catalog, on, map_ids, how, compare_type, compare_kwargs, open_kwargs, agreement_map_field, expected, expected_agreement_map",
    glob="compare_catalogs_no_kwargs",
)
def test_compare_catalogs_no_kwargs(
    candidate_catalog,
    benchmark_catalog,
    on,
    map_ids,
    how,
    compare_type,
    compare_kwargs,
    open_kwargs,
    agreement_map_field,
    expected,
    expected_agreement_map,
    tmp_path,
):
    """Tests for the _compare_catalogs function"""

    def prepend_tmp_path(df, field):
        """Prepend tmp_path to field in df"""
        if (field is not None) & (field in df.columns):

            def tmp_path_apply(row):
                return tmp_path / row[field]

            dask_meta = (
                {"meta": ("output", "O")} if isinstance(df, dd.DataFrame) else {}
            )
            df[field] = df.apply(tmp_path_apply, axis=1, **dask_meta)
        return df

    # apply tmp_path to agreement_map_field in all catalogs
    candidate_catalog = prepend_tmp_path(candidate_catalog, agreement_map_field)
    benchmark_catalog = prepend_tmp_path(benchmark_catalog, agreement_map_field)
    expected = prepend_tmp_path(expected, agreement_map_field)

    agreement_catalog = catalog_compare(
        candidate_catalog,
        benchmark_catalog,
        on=on,
        map_ids=map_ids,
        how=how,
        compare_type=compare_type,
        compare_kwargs=compare_kwargs,
        open_kwargs=open_kwargs,
        agreement_map_field=agreement_map_field,
    )

    assert isinstance(agreement_catalog, dd.DataFrame) or isinstance(
        agreement_catalog, pd.DataFrame
    )

    # load agreement maps and check metadata
    if agreement_map_field is not None:
        # load agreement maps with apply and check that they are the same

        def load_agreement_and_check(row, counter=[0]):
            """Load agreement map and check that it is the same"""
            agreement_map = rxr.open_rasterio(row[agreement_map_field], **open_kwargs)

            if expected_agreement_map[counter[0]] is not None:
                expected_agreement_map_xr = rxr.open_rasterio(
                    expected_agreement_map[counter[0]], **open_kwargs
                )
                assert isinstance(agreement_map, (xr.DataArray, xr.Dataset))
                assert isinstance(expected_agreement_map_xr, (xr.DataArray, xr.Dataset))


@parametrize_with_cases(
    "candidate_catalog, benchmark_catalog, on, map_ids, how, compare_type, compare_kwargs, open_kwargs, agreement_map_field, expected_exception",
    glob="compare_catalogs_fail",
)
def test_compare_catalogs_fail(
    candidate_catalog,
    benchmark_catalog,
    on,
    map_ids,
    how,
    compare_type,
    compare_kwargs,
    open_kwargs,
    agreement_map_field,
    expected_exception,
    tmp_path,
):
    """Tests for the compare_catalogs function"""

    with raises(expected_exception):
        catalog_compare(
            candidate_catalog,
            benchmark_catalog,
            on=on,
            map_ids=map_ids,
            how=how,
            compare_type=compare_type,
            compare_kwargs=compare_kwargs,
            open_kwargs=open_kwargs,
            agreement_map_field=agreement_map_field,
        )


# This test needs to be re-evaluated
# @parametrize_with_cases(
#     "url, collection, times, bbox, assets, allow_list, block_list, expected_catalog_df",
#     glob="stac_catalog_comparison_success",
# )
# def test_stac_catalog_comparison_success(
#     url, collection, times, bbox, assets, allow_list, block_list, expected_catalog_df
# ):
#     catalog = pystac_client.Client.open(url)
#
#     candidate_items = catalog.search(
#         datetime=times[0],
#         collections=[collection],
#         bbox=bbox,
#     ).item_collection()
#
#     candidate_catalog = stac_to_df(
#         stac_items=candidate_items,
#         assets=assets,
#         attribute_allow_list=allow_list,
#         attribute_block_list=block_list,
#     )
#
#     benchmark_items = catalog.search(
#         datetime=times[1],
#         collections=[collection],
#         bbox=bbox,
#     ).item_collection()
#
#     benchmark_catalog = stac_to_df(
#         stac_items=benchmark_items,
#         assets=assets,
#         attribute_allow_list=allow_list,
#         attribute_block_list=block_list,
#     )
#
#     arguments = {
#         "candidate_catalog": candidate_catalog,
#         "benchmark_catalog": benchmark_catalog,
#         "on": "compare_id",
#         "map_ids": "map_id",
#         "how": "inner",
#         "compare_type": "continuous",
#         "compare_kwargs": {
#             "metrics": (
#                 "coefficient_of_determination",
#                 "mean_absolute_error",
#                 "mean_absolute_percentage_error",
#             ),
#             "encode_nodata": True,
#             "nodata": -9999,
#         },
#         "open_kwargs": {"mask_and_scale": True, "masked": True},
#     }
#
#     stac_clog = catalog_compare(**arguments)
#
#     pd.testing.assert_frame_equal(
#         stac_clog,
#         expected_catalog_df,
#         check_dtype=True,
#         check_index_type=False,
#         check_like=True,
#     ), "Computed catalog did not match the expected catalog df"


@parametrize_with_cases(
    "url, collection, time, bbox, assets, allow_list, block_list, exception",
    glob="stac_catalog_comparison_fail",
)
def test_stac_catalog_comparison_fail(
    url, collection, time, bbox, assets, allow_list, block_list, exception
):
    with raises(exception):
        catalog = pystac_client.Client.open(url)

        candidate_items = catalog.search(
            datetime=time,
            collections=[collection],
            bbox=bbox,
        ).item_collection()

        _ = stac_to_df(
            stac_items=candidate_items,
            assets=assets,
            attribute_allow_list=allow_list,
            attribute_block_list=block_list,
        )
